local _G = _G
local ipairs = _G.ipairs
local UIParent = _G.UIParent
local GetGuildInfo = _G.GetGuildInfo
local UnitLevel = _G.UnitLevel
local UnitRace = _G.UnitRace
local UnitClass = _G.UnitClass
local UnitFactionGroup = _G.UnitFactionGroup
local UnitHealth = _G.UnitHealth
local UnitHealthMax = _G.UnitHealthMax
local UnitMana = _G.UnitMana
local UnitManaMax = _G.UnitManaMax
local UnitIsUnit = _G.UnitIsUnit
local UnitExists = _G.UnitExists
local UnitName = _G.UnitName
local GetTime = _G.GetTime
local RAID_CLASS_COLORS = _G.RAID_CLASS_COLORS
local GetMouseFocus = _G.GetMouseFocus
local UnitGUID = _G.UnitGUID

local font = "Interface\\AddOns\\oUF_Mouseover\\media\\font.ttf"
local fontHeight = 12
local texture = "Interface\\AddOns\\oUF_Mouseover\\media\\statusbar"

if LibStub then LibMobHealth = LibStub("LibMobHealth-4.0", true) end

local tooltip = CreateFrame("GameTooltip")
tooltip:SetOwner(_G.UIParent, "ANCHOR_NONE")
tooltip.left = {}
tooltip.right = {}
for i = 1, 30 do
	tooltip.left[i] = tooltip:CreateFontString()
	tooltip.left[i]:SetFontObject(GameFontNormal)
	tooltip.right[i] = tooltip:CreateFontString()
	tooltip.right[i]:SetFontObject(GameFontNormal)
	tooltip:AddFontStrings(tooltip.left[i], tooltip.right[i])
end

local timeout = 0
local oldName
local function updateTooltip(unit)
	local name = UnitName(unit)
	local timeNow = GetTime()
	if oldName == name then
		return
	end
	oldName = name
	timeout = timeNow + 1
	tooltip:ClearLines()
	tooltip:SetUnit(unit)
end

local function unitZone(unit)
	if not UnitIsPlayer(unit) then return end
	updateTooltip(unit)
	local lastText = tooltip.left[tooltip:NumLines()]:GetText()
	if lastText == PVP or lastText:find("^Level") then return end 
	return lastText
end

local function unitName(unit)
	updateTooltip(unit)
	return tooltip.left[1]:GetText()
end

local unitRace = function(unit)
	local race
	if UnitIsPlayer(unit) then
		race = UnitRace(unit);
	else
		race = UnitCreatureFamily(unit) or UnitCreatureType(unit)
	end
	return race
end

local unitClass = function(unit)
	local class = UnitClass(unit)
	if class == UnitName(unit) then return end
	return class
end

local getMouseoverTargetName = function(unit)
	return UnitName(unit.."target") or "None"
end

local unitMana = function(unit)
	local mana = UnitMana(unit)
	local maxMana = UnitManaMax(unit)
	local value
	if maxMana == 100 then
		value = mana
	elseif maxMana ~= 0 then
		value = format("%d/%d", mana, maxMana)
	end
	return value
end

local unitHealth = function(unit)
	local health, maxHealth
    if LibMobHealth then 
		health, maxHealth = LibMobHealth:GetUnitHealth(unit)
	else 
		health = UnitHealth(unit)
		maxHealth = UnitHealthMax(unit)
	end
	
	local value
	if maxHealth == 100 then 
		value = health .. "%"
	elseif maxHealth ~= 0 then
		value = format("%d/%d", health, maxHealth)
	end
	return value
end

local NUM_LINES = 10
local lines = {
	{"Target:", getMouseoverTargetName},
	{"Guild:", GetGuildInfo, 1},
	{"Rank:", GetGuildInfo, 2},
	{"Level:", UnitLevel},
	{"Race:", unitRace},
	{"Class:", unitClass},
	{"Faction:", UnitFactionGroup},
	{"Health:", unitHealth},
	{"Mana:", unitMana},
	{"Zone:", unitZone},
}

local powers = {
	["WARRIOR"] = "Rage:",
	["ROGUE"] = "Energy:",
}

powers = setmetatable(powers, {__index=function(self,key)
	if type(key) == nil then return nil end
	if rawget(self,key) then
		return self[key]
	else
		return "Mana:"
	end
end})

function oUF:UNIT_NAME_UPDATE(event, unit)
	if(self.unit ~= unit or not UnitExists(unit)) then return end
	local _, realm = UnitName(unit)
	local name = unitName(unit)
	
	local class = select(2,UnitClass(unit))
	local c = RAID_CLASS_COLORS[class]
	if c then 
		self.Name:SetVertexColor(c.r, c.g, c.b) 
	else 
		self.Name:SetVertexColor(1,1,1) 
	end

	for i=1, NUM_LINES do
			self["line" .. i .. "left"]:SetText()
			self["line" .. i .. "right"]:SetText()
			self["line" .. i .. "right"]:SetVertexColor(1, 1, 1)
			self["line" .. i .. "left"]:ClearAllPoints()
			self["line" .. i .. "right"]:ClearAllPoints()
	end
	
	if realm then
		self.Name:SetFormattedText("%s-%s", name, realm)
	else
		self.Name:SetText(name)
	end
	
	local fStringWidth = self.Name:GetStringWidth()
	if fStringWidth > self:GetWidth() - 8 then
		local text = self.Name:GetText()
		local fStringTextLen = text:len()
		local avgCharWidth = fStringWidth / fStringTextLen
		local charsAllowed = (self:GetWidth() - 8) / avgCharWidth
		local charsTooMany = fStringTextLen - charsAllowed
		local stringToShear = text:sub(charsTooMany * -1 - 3, text:len())
		stringToShear = stringToShear:gsub("([%^%$%(%)%%%.%[%]%*%+%-%?])", "%%%1")
		text = text:gsub(stringToShear .. "$", "") .. "..."
		self.Name:SetText(text)
	end

			
	local lineNum = 0
	if not self.inCombat then self:SetHeight(30) end
	for i,v in ipairs(lines) do
		local value
		local power
		if v[3] then
			value = select(v[3], v[2](unit))
		else
			value = v[2](unit)
		end
		if value or self.inCombat then
			lineNum = lineNum + 1
			if v[1] == "Target:" then
				if value == nil then 
					self["line" .. lineNum .. "right"]:SetVertexColor(1, 1, 1)
				else
					class = select(2, UnitClass("mouseovertarget"))
					c = RAID_CLASS_COLORS[class]
					if c then
						self["line" .. lineNum .. "right"]:SetVertexColor(c.r, c.g, c.b)
					else
						self["line" .. lineNum .. "right"]:SetVertexColor(1, 1, 1)
					end
				end
			elseif v[1] == "Mana:" then
				class = select(2, UnitClass(unit))
				power = powers[class]
			elseif v[1] == "Guild:" then
				value = "<" .. (value or "None") .. ">"
			end
			self["line" .. lineNum .. "left"]:SetText(power or v[1])
			self["line" .. lineNum .. "right"]:SetText(value or "None")			
			self["line" .. lineNum .. "left"]:SetPoint("LEFT", 2, -12+(-1 * lineNum * 12))
			self["line" .. lineNum .. "right"]:SetPoint("RIGHT", -2, -12+(-1 * lineNum * 12))
			if not self.inCombat then self:SetHeight(30 + lineNum * 12) self:Show() end
		end
	end
end

local possiblePowers = {["Mana:"] = true, ["Power:"] = true, ["Rage:"] = true}
local function updateChangingLines(self)
	for i = 1, NUM_LINES do
		lineLeft = self["line" .. i .. "left"]
		lineRight = self["line" .. i .. "right"]
		if lineLeft then
			text = lineLeft:GetText()
			if text == "Health:" then
				local newText = unitHealth("mouseover")
				if newText ~= lineRight:GetText() then
					lineRight:SetText(newText)
				end
			elseif possiblePowers[text] then
				local newText = unitMana("mouseover")
				if newText ~= lineRight:GetText() then
					lineRight:SetText(newText)
				end
			elseif text == "Target:" then
				local newText = getMouseoverTargetName("mouseover")
				if newText ~= lineRight:GetText() then
					class = select(2, UnitClass("mouseovertarget"))
					c = RAID_CLASS_COLORS[class]
					if c then
						self["line" .. i .. "right"]:SetVertexColor(c.r, c.g, c.b)
					else
						self["line" .. i .. "right"]:SetVertexColor(1, 1, 1)
					end
					lineRight:SetText(newText)
				end
				
			end
		end	
	end
end

local func = function(settings, self, unit)
	self:SetBackdrop{
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		--edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		--edgeSize=16, 
		tile = true, tileSize = 16,
		insets = {left = 1.5, right = 1.5, top = 1.5, bottom = 1.5}
		}
	self:SetBackdropColor(0,0,0,1)
	
	-- Healthbar
	local hp = CreateFrame"StatusBar"
	hp:SetStatusBarTexture(texture)
    hp:SetParent(self)
	hp:SetPoint"TOP"
	hp:SetPoint("LEFT", 1.5, 0)
	hp:SetPoint("RIGHT", -1.5, 0)
	hp:SetHeight(6)
	self.Health = hp
	
	-- Healthbar background
	local hpbg = hp:CreateTexture(nil, "BORDER")
	hpbg:SetAllPoints(hp)
	hpbg:SetTexture(texture)
	hpbg:SetBlendMode"BLEND"
	hp.bg = hpbg
	
	-- Power Bar (manabar)
	local pb = CreateFrame"StatusBar"
	pb:SetHeight(6)
	pb:SetStatusBarTexture(texture)
	pb:SetParent(self)
	pb:SetPoint("LEFT", 1.5, 0)
	pb:SetPoint("RIGHT", -1.5, 0)
	pb:SetPoint"BOTTOM"
	self.Power = pb
	
	-- Powerbar background
	local pbbg = pb:CreateTexture(nil, "BORDER")
	pbbg:SetAllPoints(pb)
	pbbg:SetBlendMode"BLEND"
	pbbg:SetAlpha(0.3)
	pbbg:SetTexture(.25,.25,.25)
	pb.bg = pbbg
	    
	-- Raid target icons
    local icon = hp:CreateTexture(nil, "OVERLAY")
    icon:SetHeight(16)
    icon:SetWidth(16)
    icon:SetPoint("TOP", self, 0, 8)
    icon:SetTexture"Interface\\TargetingFrame\\UI-RaidTargetingIcons"
    self.RaidIcon = icon
	
    -- Name
    local name = hp:CreateFontString(nil, "OVERLAY")
    name:SetPoint("LEFT", 2, -12)
    name:SetJustifyH"LEFT"
	name:SetFont(font, fontHeight)
    name:SetFontObject(GameFontNormal)
    name:SetTextColor(1,1,1)
    self.Name = name

	-- Lines
	for i=1,NUM_LINES do
		local lineLeft = hp:CreateFontString(nil, "OVERLAY")
		lineLeft:SetFontObject(GameFontNormal)
		lineLeft:SetFont(font, fontHeight)
		lineLeft:SetTextColor(1,1,1)
		lineLeft:SetJustifyH"LEFT"
		self["line" .. i .. "left"] = lineLeft
		local lineRight = hp:CreateFontString(nil, "OVERLAY")
		lineRight:SetFontObject(GameFontNormal)
		lineRight:SetFont(font, fontHeight)
		lineRight:SetTextColor(1,1,1)
		lineRight:SetJustifyH"RIGHT"
		self["line" .. i .. "right"] = lineRight
	end
	
	-- PVP
	local pvp = hp:CreateTexture(nil, "OVERLAY")
	pvp:SetHeight(30)
	pvp:SetWidth(30)
	pvp:SetPoint("TOPRIGHT", self, 20, 10)
	self.PvP = pvp
	
	self.ignoreHealComm = true
end

oUF:RegisterStyle("Mouseover", setmetatable({
["initial-width"] = 180,
["initial-height"] = 30
}, {__call = func}))

oUF:SetActiveStyle"Mouseover"
local mouseover = oUF:Spawn"mouseover"
mouseover:SetFrameStrata"TOOLTIP"
mouseover:SetPoint("BOTTOMRIGHT", UIParent, "BOTTOMRIGHT", -12, 12)
if mouseover:IsEventRegistered"UPDATE_MOUSEOVER_UNIT" then
	mouseover:UnregisterEvent"UPDATE_MOUSEOVER_UNIT"
end

local frame = CreateFrame("Frame", nil, mouseover)
local oldX, oldY = 0, 0
local oldGUID
frame.timeout1, frame.timeout2, frame.timeout3 = 0, 0, 0
frame:SetScript("OnUpdate", function(self, update)
	self.timeout1 = self.timeout1 + update
	if self.timeout1 > .05 then
		local frame = self:GetParent()	
		if frame:IsShown() then
			if not frame.inCombat then
				local x, y = GetCursorPosition()
				local effScale = frame:GetEffectiveScale()
				local mouseFocus = GetMouseFocus()
				if mouseFocus and mouseFocus.unit and mouseFocus.unit ~= "mouseover" then
					frame:ClearAllPoints()
					frame:SetPoint("BOTTOMRIGHT", UIParent, "BOTTOMRIGHT", -12, 12)
				elseif x ~= oldX or y ~= oldY then
					frame:ClearAllPoints()
					frame:SetPoint("BOTTOM",UIParent,"BOTTOMLEFT",(x / effScale),(y / effScale) + 6)
				end
				oldX, oldY = x, y
			end
		end
		self.timeout1 = 0
	end
	self.timeout2 = self.timeout2 + update
	if self.timeout2 > .5 then
		local frame = self:GetParent()
		if frame:IsShown() then
			updateChangingLines(frame)
			frame:UNIT_HEALTH(nil, frame.unit)
			frame:UNIT_MANA(nil, frame.unit)
		end
		self.timeout2 = 0
	end
	self.timeout3 = self.timeout3 + update
	if self.timeout3 > .25 then
		local frame = self:GetParent()
		if frame:IsShown() then
			local unitGUID = UnitGUID("mouseover")
			if unitGUID ~= oldGUID then
				frame:UPDATE_MOUSEOVER_UNIT()
				oldGUID = unitGUID
			end
		end
		self.timeout3 = 0
	end
end)

frame:RegisterEvent("PLAYER_REGEN_DISABLED")
frame:RegisterEvent("PLAYER_REGEN_ENABLED")
frame:SetScript("OnEvent", function(self, event, ...)
	self = self:GetParent()
	if event == "PLAYER_REGEN_DISABLED" then
		self:SetHeight(30 + NUM_LINES * 12)
		self:ClearAllPoints()
		self:SetPoint("BOTTOMRIGHT", UIParent, "BOTTOMRIGHT", -12, 12)
		self.inCombat = true
	elseif event == "PLAYER_REGEN_ENABLED" then
		self.inCombat = false
	end
end)

GameTooltip:SetScript("OnTooltipSetUnit", function(tooltip) tooltip:Hide() end )